#!/usr/bin/env python

import numpy as np
from sklearn.linear_model import LinearRegression, HuberRegressor, Lasso
from skimage.transform import rescale
import cv2

import sys
from sys import argv 
import os
from os.path import splitext, exists, basename, join
from os import remove, makedirs, listdir, getcwd, chdir, environ

import glob
import shutil
import iio
import rasterio

import matplotlib.pyplot as plt
from matplotlib import cm

from scipy import special
import math

def log(a):
    return np.log(np.maximum(a, 1e-64))

robust = False
use_true_ratio = True
patch_rec = False
patch_det = False
ps=30
Latm = 1.6

SCALE = 0.0005

# Read images
in_dir = argv[argv.index("--im-dir") + 1]
im_paths = sorted(glob.glob(join(in_dir, "*.tif")))
# Index of the ref image
index = int(argv[argv.index("--index") + 1])
index = min(len(im_paths), index)
im_paths = im_paths[:index]
#ims = [iio.read(path) for path in im_paths]
# Read image and remove aliasing
ims = [cv2.GaussianBlur(rasterio.open(path).read().transpose((1,2,0)).astype(np.float), (0,0), 1) for path in im_paths]
#ims = [rasterio.open(path).read().transpose((1,2,0)).astype(np.float) for path in im_paths]
shape = ims[0][:,:,0].shape

# Cloud detection method to use (can be combined)
cloudThresh = int(argv[argv.index("--cloudThresh") + 1])/100
cd1 = (str(argv[argv.index("--cd1") + 1]).lower() == 'true') # Dagobert disparity
cd2 = (str(argv[argv.index("--cd2") + 1]).lower() == 'true') # Dagobert timeseries
cd3 = (str(argv[argv.index("--cd3") + 1]).lower() == 'true') # Grompone disparity
cd4 = (str(argv[argv.index("--cd4") + 1]).lower() == 'true') # Grompone timeseries
cds = [cd1, cd2, cd3, cd4]
if not cd1 and not cd2 and not cd3 and not cd4:
    # Don't use any cloud detector, keep all images
    cloudCov = [True for im in ims]
    clouds = [np.zeros(shape) for i in range(len(ims))]
else:
    clouds = []
    # Combine the detections from selected cloud detector
    for i in range(4):
        if cds[i] and exists(join(in_dir, str(i))):
            cloud_paths = sorted(glob.glob(join(in_dir, str(i) + "/masks/*.png")))
            if len(clouds) == 0:
                clouds = [iio.read(path)[:,:,0] > 0 for path in cloud_paths]
            else:
                clouds = [np.logical_or(cloud, iio.read(path)[:,:,0] > 0) for cloud, path in zip(clouds, cloud_paths)]

    if len(clouds) == 0:
        cloudCov = [True for im in ims]
        clouds = [np.zeros(shape) for i in range(len(ims))]

    # Compute the cloud coverage for each image
    npixels = np.prod(shape)
    cloudCov = [np.sum(cloud.flatten())/npixels < cloudThresh for cloud in clouds]
    cloudCov = cloudCov[:index]
    # Always keep the reference image
    cloudCov[-1] = True

# Indices of the bands to use
mBand = int(argv[argv.index("--mb") + 1])
nomBand = int(argv[argv.index("--nomb") + 1])
ratioBand = int(argv[argv.index("--ratiob") + 1])

# Process and print the data
algo_info = open("algo_info.txt","w")
algo_info.write('nim='+str(index)+'\n')
for i in range(index):
    algo_info.write('date_'+str(i)+'='+basename(im_paths[i])[:10]+'\n')
algo_info.close()

# Rescale everything and save the visualization 
resc = np.stack(log(ims), 0)
maxi = np.percentile(resc, 99, axis=(0,1,2), keepdims=True)
mini = np.percentile(resc,  1, axis=(0,1,2), keepdims=True)
resc -= mini
resc /= (maxi - mini)
resc *= 255
# Write three bands
[iio.write("mb_"+str(i)+".png", im[:,:,mBand]) for i, im in enumerate(resc)]
[iio.write("nomb_"+str(i)+".png", im[:,:,nomBand]) for i, im in enumerate(resc)]
[iio.write("ratiob_"+str(i)+".png", im[:,:,ratioBand]) for i, im in enumerate(resc)]
# Save the clouds
for i, (cloud, cov) in enumerate(zip(clouds, cloudCov)):
    cloudrgb = np.zeros((shape[0], shape[1], 3))
    if cov:
        cloudrgb[:,:,1] = cloud*255
    else:
        cloudrgb[:,:,0] = cloud*255
    iio.write("cloud_"+str(i)+".png", cloudrgb)
# Save ratios
if use_true_ratio:
    trbm = [log(im[:,:,mBand])-log(im[:,:,ratioBand]) for im in ims]
else:
    trbm = [log(im[:,:,mBand])-log(np.mean(im[:,:,ratioBand])) for im in ims]
trbm = np.stack(trbm, 0)
maxi = np.percentile(trbm, 99, axis=(0,1,2), keepdims=True)
mini = np.percentile(trbm,  1, axis=(0,1,2), keepdims=True)
trbm -= mini
trbm /= (maxi-mini)
trbm *= 255
[iio.write("rmb_"+str(i)+".png", im) for i, im in enumerate(trbm)]

if use_true_ratio:
    trbnom = [log(im[:,:,nomBand])-log(im[:,:,ratioBand]) for im in ims]
else:
    trbnom = [log(im[:,:,nomBand])-log(np.mean(im[:,:,ratioBand])) for im in ims]
trbnom = np.stack(trbnom, 0)
maxi = np.percentile(trbnom, 99, axis=(0,1,2), keepdims=True)
mini = np.percentile(trbnom,  1, axis=(0,1,2), keepdims=True)
trbnom -= mini
trbnom /= (maxi-mini)
trbnom *= 255
[iio.write("rnomb_"+str(i)+".png", im) for i, im in enumerate(trbnom)]

# Remove the images with too many cloud
ims = [im for im, cloud in zip(ims, cloudCov) if cloud]

bm = [log(im[:,:,mBand]) for im in ims]
bnom = [log(im[:,:,nomBand]) for im in ims]
if use_true_ratio:
    bratio = [log(im[:,:,ratioBand]) for im in ims]
else:
    bratio = [np.mean(log(im[:,:,ratioBand])) for im in ims]

if len(bm) == 1: # Only the reference image available
    print("The past is completely cloudy or non-existant")
    exit(1)

######################## REC ##################################

if not patch_rec:
    # TODO project using only the region without cloud
    X1 = np.asarray([cm.flatten() - cratio.flatten() for cm, cratio in zip(bm[:-1], bratio[:-1])])
    y1 = np.asarray([bm[-1].flatten() - bratio[-1].flatten()])

    X2 = np.asarray([cnom.flatten() - cratio.flatten() for cnom, cratio in zip(bnom[:-1], bratio[:-1])])
    y2 = np.asarray([bnom[-1].flatten() - bratio[-1].flatten()])

    # Linear regression/Robust linear regression
    estimator = HuberRegressor(max_iter=1000) if robust else LinearRegression()

    info = open("rec_info.txt","w") 
    info.write(''.join([f'{i+1:^10}' for i, cov in enumerate(cloudCov[:-1]) if cov]) + '\n') 

    reg = estimator.fit(X1.T, y1.T.ravel())
    pred1 = reg.predict(X1.T)
    diff1 = y1.flatten() - pred1.flatten()
    info.write(''.join([f'{i:^10.6f}' for i in reg.coef_]) + '\n') 

    reg = estimator.fit(X2.T, y2.T.ravel())
    pred2 = reg.predict(X2.T)
    diff2 = y2.flatten() - pred2.flatten()
    info.write(''.join([f'{i:^10.6f}' for i in reg.coef_]) + '\n') 
    info.close()

    diff1 = diff1.reshape(shape)
    diff2 = diff2.reshape(shape)
else:
    # Linear regression/Robust linear regression
    estimator = HuberRegressor(max_iter=1000) if robust else LinearRegression()

    diff1 = np.zeros(shape)
    diff2 = np.zeros(shape)
    #for i in range(int(math.ceil(shape[0]/ps))):
    #    for j in range(int(math.ceil(shape[1]/ps))):

    for i in range(int(math.floor(shape[0]/ps))):
        for j in range(int(math.floor(shape[1]/ps))):
            X1 = np.asarray([cm[i*ps:(i+1)*ps, j*ps:(j+1)*ps].flatten() - cratio[i*ps:(i+1)*ps, j*ps:(j+1)*ps].flatten() for cm, cratio in zip(bm[:-1], bratio[:-1])])
            y1 = np.asarray([bm[-1][i*ps:(i+1)*ps, j*ps:(j+1)*ps].flatten() - bratio[-1][i*ps:(i+1)*ps, j*ps:(j+1)*ps].flatten()])

            X2 = np.asarray([cnom[i*ps:(i+1)*ps, j*ps:(j+1)*ps].flatten() - cratio[i*ps:(i+1)*ps, j*ps:(j+1)*ps].flatten() for cnom, cratio in zip(bnom[:-1], bratio[:-1])])
            y2 = np.asarray([bnom[-1][i*ps:(i+1)*ps, j*ps:(j+1)*ps].flatten() - bratio[-1][i*ps:(i+1)*ps, j*ps:(j+1)*ps].flatten()])

            reg = estimator.fit(X1.T, y1.T.ravel())
            pred1 = reg.predict(X1.T)
            diff1[i*ps:(i+1)*ps, j*ps:(j+1)*ps] = y1.reshape((ps,ps)) - pred1.reshape((ps,ps))

            reg = estimator.fit(X2.T, y2.T.ravel())
            pred2 = reg.predict(X2.T)
            diff2[i*ps:(i+1)*ps, j*ps:(j+1)*ps] = y2.reshape((ps,ps)) - pred2.reshape((ps,ps))


################## Write the result #############################################

out_diff1 = cm.get_cmap(plt.get_cmap('coolwarm'))(diff1 / SCALE / 255 / 2 + 0.5)
iio.write("res_methane.png", out_diff1*255)

out_diff2 = cm.get_cmap(plt.get_cmap('coolwarm'))(diff2 / SCALE / 255 / 2 + 0.5)
iio.write("res_nomethane.png", out_diff2*255)

mini = np.percentile(y1, 1)
maxi = np.percentile(y1, 99)
y1 = (y1-mini)/(maxi-mini)*255
iio.write("ratio_methane.png", y1.reshape(shape))
pred1 = (pred1-mini)/(maxi-mini)*255
iio.write("rec_methane.png", pred1.reshape(shape))

mini = np.percentile(y2, 1)
maxi = np.percentile(y2, 99)
y2 = (y2-mini)/(maxi-mini)*255
iio.write("ratio_nomethane.png", y2.reshape(shape))
pred2 = (pred2-mini)/(maxi-mini)*255
iio.write("rec_nomethane.png", pred2.reshape(shape))

maxi = np.percentile(ims[-1][:,:,1:4], 99, axis=(0,1,2), keepdims=True)
rgb = ims[-1][:,:,1:4]/maxi*255
iio.write("rgb.png", rgb[:,:,::-1])

############### Automatic detections ##########################################

if not patch_det:
    sigma_x = np.sqrt(np.mean(diff1**2))
    sigma_y = np.sqrt(np.mean(diff2**2))
    rho = np.mean(diff1 * diff2) / (sigma_x * sigma_y)

    #mu = rho * sigma_x / sigma_y * diff2
    mu = - rho * sigma_x / sigma_y * np.abs(diff2) #
    sigma = sigma_x * np.sqrt(1 - rho*rho)
    print(rho, sigma)

    lognfas = np.log(np.prod(shape)) + np.log(np.maximum(special.erfc(-(diff1 - mu)/(np.sqrt(2)*sigma))/2, 1e-64))
    lognfas1 = np.log(np.prod(shape)) + np.log(np.maximum(special.erfc(-diff1/(np.sqrt(2)*sigma_x))/2, 1e-64))
    lognfas2 = np.log(np.prod(shape)) + np.log(np.maximum(special.erfc(np.abs(diff2)/(np.sqrt(2)*sigma_y))/2, 1e-64))

    lognfathresh = float(argv[argv.index("--lognfa") + 1])
    iio.write("detect.png", (np.logical_and(lognfas < lognfathresh, np.logical_not(lognfas2 < lognfathresh))*255).reshape(shape))
    iio.write("detect1.png", ((lognfas1 < lognfathresh)*255).reshape(shape))
    iio.write("detect2.png", ((lognfas2 < lognfathresh)*255).reshape(shape))
else:
    lognfas = np.zeros(diff1.shape)
    lognfas1 = np.zeros(diff1.shape)
    lognfas2 = np.zeros(diff1.shape)

    for i in range(int(math.ceil(shape[0]/ps))):
        for j in range(int(math.ceil(shape[1]/ps))):
            _diff1 = diff1[i*ps:(i+1)*ps, j*ps:(j+1)*ps]
            _diff2 = diff2[i*ps:(i+1)*ps, j*ps:(j+1)*ps]

            sigma_x = np.sqrt(np.mean(_diff1**2))
            sigma_y = np.sqrt(np.mean(_diff2**2))
            rho = np.mean(_diff1 * _diff2) / (sigma_x * sigma_y)

            #mu = rho * sigma_x / sigma_y * diff2
            mu = - rho * sigma_x / sigma_y * np.abs(_diff2) #
            sigma = sigma_x * np.sqrt(1 - rho*rho)
            print(rho, sigma, sigma_x, sigma_y)

            lognfas[i*ps:(i+1)*ps, j*ps:(j+1)*ps] = np.log(np.prod(shape)) + np.log(np.maximum(special.erfc(-(_diff1 - mu)/(np.sqrt(2)*sigma))/2, 1e-64))
            lognfas1[i*ps:(i+1)*ps, j*ps:(j+1)*ps] = np.log(np.prod(shape)) + np.log(np.maximum(special.erfc(-_diff1/(np.sqrt(2)*sigma_x))/2, 1e-64))
            lognfas2[i*ps:(i+1)*ps, j*ps:(j+1)*ps] = np.log(np.prod(shape)) + np.log(np.maximum(special.erfc(np.abs(_diff2)/(np.sqrt(2)*sigma_y))/2, 1e-64))

    lognfathresh = float(argv[argv.index("--lognfa") + 1])
    iio.write("detect.png", (np.logical_and(lognfas < lognfathresh, np.logical_not(lognfas2 < lognfathresh))*255).reshape(shape))
    iio.write("detect1.png", ((lognfas1 < lognfathresh)*255).reshape(shape))
    iio.write("detect2.png", ((lognfas2 < lognfathresh)*255).reshape(shape))

##################################### TEMP ####################################

#iio.write("nfas.tiff", lognfas.reshape(shape))
##iio.write("diff.tiff", (diff1-mu).reshape(shape))
#iio.write("diff1.tiff", diff1)
#iio.write("diff2.tiff", diff2)

############### Methane concentration #############################################
#
#from scipy.optimize import fmin
#from scipy import interpolate
#
#data = np.loadtxt('SpectrMol_77d19c7f8a93c86d8250ba5a160145ee951bbc34.coef.txt')
#absor = data[:,1]
#wavelenght = 10000000/data[:,0]
#s2 = np.loadtxt('s2_response_freq.csv',skiprows=1,delimiter=',')
#
#f = interpolate.interp1d(s2[:,0], s2[:,13])
#sensitivity = f(wavelenght)
#
#alpha = np.exp(-sigma*np.sqrt(2)*special.erfcinv(2*np.exp(lognfathresh)/np.prod(shape))/2) 
## Formula per pixel
##alpha = np.exp(np.minimum(diff1 - mu, 0)) 
#
#def g(L):
#    #return (sensitivity.dot(np.exp(-2*absor*L)) / np.sum(sensitivity)) # Model with fully transparent atmosphere
#    return (sensitivity.dot(np.exp(-2*absor*(L+Latm))) /  sensitivity.dot(np.exp(-2*absor*Latm))) # Model with an atmosphere made of only methane
#
#def h(L):
#    return (alpha - g(L) )**2
#
## This optimization is precise but slow. If it needs to be done multiple time (for example for each pixels),
## the best its to either tabulate this function or invert a Taylor expansion (see Gabriele's notebook, it's
## an approximation valid for small quantities of methane)
#L = fmin(h, 0)
#
#print ("Minimum detection is {}cm of methane, which amounts to {}ppb".format(L, (L*2.5e19/2e25)*1e9 ))
